%% PREVIEW CONTROL DEMO
% Benjamin Stephens - 2009
% Robotics Institute, Carnegie Mellon University

clear all;clc
% close all;

%% Paramters
L = 1.0;
g=9.81;
T = 0.02;  % timestep

step_times = 0.2:0.02:0.5;
init_vels = 0.3:0.02:0.5;

for k = 1:length(init_vels)
for n = 1:length(step_times)
    

Tstep = step_times(n);
Nstep = floor(Tstep/T);
m = 93;
mf = 10;

T2 = Tstep-Nstep*T;

dx = 0.1;
dy = 0.05;

N = ceil(1.5/T)+1;

a = 1e-2;
b = 0.1;
c = 0.01;
d = 1e-6;
e = 1e-5;

%% initial condition

xs0 = 0;
ys0 = 0.08;

x0 = [0,init_vels(k),0,xs0,0]';
% y0 = [-0.12,0.3,0,ys0,0]';
xf0 = 0;
% yf0 = -0.12;

vxmax = 0.8*3.0;
% vymax = 1.0;

% afmax = 20.0;
ufmax = 80.0;

preview_1D_time_opt;

step_dist(n,k) = xf;
step_vals(n,k) = fval;

end
[opt_val(k),idx] = min(step_vals(:,k));
opt_time(k) = step_times(idx);
opt_dist(k) = step_dist(idx,k);


end
%%
figure(1);clf;
plot(step_times,step_dist,'LineWidth',2);
hold on;
plot(opt_time,opt_dist,'*','MarkerSize',6);
set(gca,'FontSize',12);
xlabel('Step Time(s)');
ylabel('Optimal Step Distance(m)');
figure(2);clf;
plot(step_times,step_vals,'LineWidth',2);
hold on;
plot(opt_time,opt_val,'*','MarkerSize',6);
set(gca,'FontSize',12);
xlabel('Step Time(s)');
ylabel('Optimization Value');
% ylim([min(opt_val)-0.1,min([max(opt_val) + 0.1,0.1])]);

%
coeffs = [ones(length(init_vels),1),init_vels',init_vels'.^2]\opt_time';

est_time = [ones(length(init_vels),1),init_vels',init_vels'.^2]*coeffs

figure(1);
plot(est_time,opt_dist,'g*-','LineWidth',2);

%%
figure(3);clf;
for n = 1:length(est_time)
    Tstep = est_time(n);
    Nstep = floor(Tstep/T);
    T2 = Tstep-Nstep*T;
    N = ceil(1.5/T)+1;
    x0 = [0,init_vels(n),0,xs0,0]';
    
a = 1e-2;
b = 0.1;
c = 0.01;
d = 1e-6;
e = 1e-5;    
    
    preview_1D_time_opt;
    
    Xf = Pfpx*x0 + Pfpu*ux;
    xf = Us*Xf;
    px = Ppx*x0 + Ppu*ux;
    zx = Pzx*x0 + Pzu*ux;
    figure(3); 
    plot(px,'g-'); hold on;
    plot(zx,'b-');
%     plot(Xf,'r*-');
%     legend('COM','ZMP','Swing Foot','Location','Southeast');
%     plot(xf0,'g*');
%     plot(xf,'g*');        
end
    



%%
% % uxy = reshape(u,2*(Nstep+1) + N-Nstep-1,2);
% 
% % ux = uxy(:,1);
% % uy = uxy(:,2);
% 
% Xf = Pfpx*x0 + Pfpu*ux;
% 
% xf = Us*Xf;
% 
% % dp = [xf-xf0;yf-yf0];
% % 
% px = Ppx*x0 + Ppu*ux;
% % vx = Pvx*x0 + Pvu*ux;
% % vy = Pvx*y0 + Pvu*uy;
% zx = Pzx*x0 + Pzu*ux;
% % 
% % zxref = Uc*xf0 + 0.5*U*(xf0+xf);
% % zyref = Uc*yf0 + 0.5*U*(yf0+yf);
% % 
% % time = T:T:N*T;
% % 
% figure(1);clf;
% % 
% plot(px,'g-'); hold on;
% plot(zx,'b-');
% % % plot(zxref,zyref,'k--');
% plot(Xf,'r*-');
% legend('COM','ZMP','Swing Foot','Location','Southeast');
% plot(xf0,'g*');
% plot(xf,'g*');
% % plot(0.5*(xf0+xf),0.5*(yf0+yf),'go','LineWidth',3);
% % % plot([xf0;xf0+0.2*(x0(1)-xf0+x0(2)*sqrt(L/g))],[yf0;yf0+0.2*(y0(1)-yf0+y0(2)*sqrt(L/g))],'m-');
% % 
% % yf0 = 0;
% % ys0 = 0;
% % h = plot([xf0-dx,xf0+dx,xf0+dx,xf0-dx,xf0-dx]',[yf0-dy,yf0-dy,yf0+dy,yf0+dy,yf0-dy]','k-',...
% %          [xs0-dx,xs0+dx,xs0+dx,xs0-dx,xs0-dx]',[ys0-dy,ys0-dy,ys0+dy,ys0+dy,ys0-dy]','k--',...
% %          [xf-dx,xf+dx,xf+dx,xf-dx,xf-dx]',[yf-dy,yf-dy,yf+dy,yf+dy,yf-dy]','m-');
% % xlabel('x');
% % ylabel('y');
%      
% % uxy0 = reshape(u0,N+1,2);
% % xfu = uxy0(N+1:end,1);
% % yfu = uxy0(N+1:end,2);     
% % h = plot([xfu-dx,xfu+dx,xfu+dx,xfu-dx,xfu-dx]',[yfu-dy,yfu-dy,yfu+dy,yfu+dy,yfu-dy]','m--');
% % h = plot([z10(1),z20(1);z10(1)+dp0(1),z20(1)+dp0(1)],[z10(2),z20(2);z10(2)+dp0(2),z20(2)+dp0(2)],'r--');
% % h = plot([z1(1),z2(1);z1(1)+dp(1),z2(1)+dp(1)],[z1(2),z2(2);z1(2)+dp(2),z2(2)+dp(2)],'r-');
% % 
% % axis equal;
% % 
% % figure(2);clf;
% % subplot 311
% % plot(Xf-xs0,'b');
% % hold on;
% % plot(Yf-ys0,'r');
% % legend('x','y');
% % ylabel('position');
% % title('Swing Foot Trajectory');
% % subplot 312
% % plot(Pfvx*x0+Pfvu*ux,'b');
% % hold on;
% % plot(Pfvx*y0+Pfvu*uy,'r');
% % ylabel('velocity');
% % subplot 313
% % plot(Uf*ux,'b');
% % hold on;
% % plot(Uf*uy,'r');
% % ylabel('force');
% % xlabel('timesteps');
% % hold on;
% % plot(Yf);
% 
% % gamma = 0.2;
% % beta = (1+gamma)/(1-gamma);
% % k=(1+beta)/sqrt(beta^2-1);
% % ys = k*sqrt(L/g*y0(2)^2-(y0(1)-yf0)^2)
% % xs = k*sqrt(L/g*x0(2)^2-(x0(1)-xf0)^2)
% 
% % 
% % figure(2);clf;
% % subplot 311
% % plot(time,px,time,py);
% % ylabel('position');
% % legend('x','y','Location','Southeast');
% % subplot 312
% % plot(time,vx,time,vy);
% % ylabel('velocity');
% % subplot 313
% % plot(time,zx,time,zy);
% % ylabel('zmp');
% 
% % plot(time,Zref,'k--'); hold on;
% % plot(time,z,'r-');
% % plot(time,p,'b-');
% % legend('zmp ref','zmp','com','location','southeast');
% % xlabel('time (s)');
% % title('preview control');

