%% Linear system lookahead
% x1 = A*x0 + B*u1
% x2 = A*x1 + B*u1 = A^2*x0 +   A*B*u0 +   B*u2
% x3 = A*x2 + B*u2 = A^3*x0 + A^2*B*u0 + A*B*u1 + B*u3
% ... iterate ...
% xN = ...         = A^N*x0 + A^(N-1)*B*u0 + ... + B*uN

bf = g/L*m/(m-mf);

A1 = [1,T,T*T*0.5,0,0;
     0,1,T,0,0;
     0,0,1,0,0;
     0,0,0,1,T;
     T*bf,0,0,-T*bf,1];
B1 = [T*T*T/6,0;T*T/2,0;T,0;0,T^2/2/mf;0,T/mf];

A2 = [1,T2,T2*T2*0.5,0,0;
     0,1,T2,0,0;
     0,0,1,0,0;
     0,0,0,1,T2;
     T2*bf,0,0,-T2*bf,1];
B2 = [T2*T2*T2/6,0;T2*T2/2,0;T2,0;0,T2^2/2/mf;0,T2/mf];

T3 = T-T2;
A3 = [1,T3,T3*T3*0.5,0,0;
     0,1,T3,0,0;
     0,0,1,0,0;
     0,0,0,1,0;
     0,0,0,0,1];
B3 = [T3*T3*T3/6;T3*T3/2;T3;0;0];

A4 = [1,T,T*T*0.5,0,0;
     0,1,T,0,0;
     0,0,1,0,0;
     0,0,0,1,0;
     0,0,0,0,1];
B4 = [T*T*T/6;T*T/2;T;0;0];

C = [1,0,-L/g,0,0];
%%
% P matrix during swing
% 2*(Nstep+1) + 1*(N-Nstep-1)

P = [A1,B1,zeros(5,2*(Nstep+1) + 1*(N-Nstep-1)-2 )];
Z = C*P;
Uu = [1,0];
Uf = [0,1];
for i=1:N-1
%     i
    if (i<Nstep+1)  % swing phase
        if (i==Nstep)            
            Pi = A2*P(end-4:end,1:end);
            Pi(:,5 + 2*i + (1:2)) = B2;
        else
            Pi = A1*P(end-4:end,1:end);
            Pi(:,5 + 2*i + (1:2)) = B1;
        end  
        
        Uu = blkdiag(Uu,[1,0]);
        Uf = blkdiag(Uf,[0,1]);
        
    else % after touchdown
        if (i==Nstep+1)
            Pi = A3*P(end-4:end,1:end);
            Pi(:,5 + 2*(Nstep+1) + i - Nstep) = B3;
        else
            Pi = A4*P(end-4:end,1:end);
            Pi(:,5 + 2*(Nstep+1) + i - Nstep) = B4;
        end  
        Uu = blkdiag(Uu,1);
        Uf = blkdiag(Uf,0);
    end
%     size(Pi)
    P = [P;Pi];
    Z = [Z;C*Pi];

end

% positions
Ppx = P(1:5:end,1:5);
Ppu = P(1:5:end,6:end);
% velocities
Pvx = P(2:5:end,1:5);
Pvu = P(2:5:end,6:end);
% accelerations
Pax = P(3:5:end,1:5);
Pau = P(3:5:end,6:end);
% foot position
Pfpx = P(4:5:end,1:5);
Pfpu = P(4:5:end,6:end);
% foot velocity
Pfvx = P(5:5:end,1:5);
Pfvu = P(5:5:end,6:end);
% % foot acceleration
Pfax = -bf*(Pfpx-Ppx);
Pfau = Uf - bf*(Pfpu-Ppu);
%
% zmp
Pzx = Z(:,1:5);
Pzu = Z(:,6:end);

% step selection matrices
% Us0 = [zeros(N,3),ones(N,1),zeros(N,1)];
% Us1 = [zeros(N,Nstep+1),[zeros(Nstep+1,N-Nstep-1);eye(N-Nstep-1)]];
Us = [zeros(1,Nstep+1),1,zeros(1,N-Nstep-2)];

U0 = [ones(Nstep+1,1);zeros(N-Nstep-1,1)];
Un = ones(N,1);
U1 = 1-U0;



%% cost function
% J = 0.5*a*(Zref-Z)'*(Zref-Z) 
%   + 0.5*b*(Pref-P)'*(Pref-P) 
%   + 0.5*c*(Vref-V)'*(Vref-V) 
%   + 0.5*d*u'*u
%   + 0.5*e*uf'*uf
%
% Z = Pzx*x0 + Pzu*u
% P = Ppx*x0 + Ppu*u
% V = Pvx*x0 + Pvu*u
%
% % Swing leg trajectory, Xs0 = initial position of swing/stepping foot
% Ps = Pfpx*x0 + Pfpu*u
%
% Modify Zref for foot placement, Xf0 = position of stance foot
%
%   Zref = U0*Xf0 + U1*0.5*(Ps + Xf0)
%   Pref = 0.5*U1*(Ps + Xf0)
%
%
%% Use the symbolic math toolbox to derive the math:
% clear all;clc;
% syms a b c d e U0 Un U1 Us Xf0 Xf Pzu Pzx Ppu Ppx Pvx Pvu u Uu Uf x0 Pfpx Pfpu real;
% Z = Pzx*x0 + Pzu*u;
% V = Pvx*x0 + Pvu*u;
% P = Ppx*x0 + Ppu*u;
% Ps = Pfpx*x0 + Pfpu*u;
% % Pref = 0.5*(Ps + Xf0);
% Pref = Un*Ps;
% Zref = U0*Xf0 + U1*Pref;
% 
% J = 0.5*a*(Zref-Z)'*(Zref-Z) + 0.5*b*(Pref-P)'*(Pref-P) + 0.5*c*V'*V + 0.5*d*u'*Uu'*Uu*u + 0.5*e*u'*Uf'*Uf*u ;
% 
% H = jacobian(jacobian(J,u),u)
% 
% f = jacobian(J-u*H*u,u);
% 
% fux = jacobian(f(1),x0)
% fuf = jacobian(f(1),Xf0)

% p = [fux,fuf,fus
%      ffx,fff,ffs]*[x0;Xfc;Xs0]
%%


H = a*(diag(U1)*Pfpu-Pzu)'*(diag(U1)*Pfpu-Pzu)...
   +b*(Pfpu-Ppu)'*(Pfpu-Ppu)...
   +c*Pvu'*Pvu + d*Uu'*Uu + e*Uf'*Uf;

fux = a*(diag(U1)*Pfpu-Pzu)'*(diag(U1)*Pfpx-Pzx)...
     +b*(Pfpu-Ppu)'*(Pfpx-Ppx)...
     +c*Pvu'*Pvx;
fuf = a*(diag(U1)*Pfpu-Pzu)'*U0;
 
p = [fux,fuf];
 
% Q = blkdiag(H,H); % for 3D
Q = H;

% p = @(x0,y0,xf0,yf0)f*[x0;xf0];         

%% optimal controller

% this is the unconstrained solution
% f = p(x0,y0,xf0,yf0);
f = p*[x0;xf0];
u0 = -Q\f;

%% ZMP constraints
% NOTE: constrained solution can be found with quadprog with inequality
% constraints on the location of the ZMP, 
% if z = Pzx*x0 + Pzu*u
% ans xf = Uc*xfc0 + U*Xf
% and xf - a < z < xf + b
% then
% -Pzu*u + U*Xf <  Pzx*x0 - Uc*xfc0 + a
%  Pzu*u - U*Xf < -Pzx*x0 + Uc*xfc0 + b
% .. which can be re-written as A*u<b as needed for quadprog
%
% Currently the constraints in double support are a box with the edges
% associated with the min/max edges of the support polygon.  This is not a
% conservative approximation, but it's better than nothing :).
%
% ax = dx;
% bx = dx;
% ay = dy;
% by = dy;

% Ax=[];
% bx=[];
% Ay=[];
% by=[];

% ZMP double support box constraints
dirx = sign(x0(1)-xf0 + x0(2)*sqrt(L/g));
% diry = sign(y0(1)-yf0 + y0(2)*sqrt(L/g));
if (dirx>0)
%     U0*(xf0 - a) + U1*(xf0 - a) < Pzx*x0 + Pzu*ux < U0*(xf0 + b) + U1*(Pfpx*x0 + Pfpu*u + b);
    Ax = [ Pzu-diag(U1)*Pfpu;
          -Pzu];
    bx = [ U0*(xf0+dx)+diag(U1)*(Pfpx*x0+dx)-Pzx*x0;
          -U0*(xf0-dx)-U1*(xf0-dx)+Pzx*x0 ];    
else    
%     U0*(xf0 - a) + U1*(Pfpx*x0 + Pfpu*u - a) < Pzx*x0 + Pzu*ux < U0*(xf0 + b) + U1*(xf0 + b);
    Ax = [ Pzu;
          -Pzu+diag(U1)*Pfpu];
    bx = [ U0*(xf0+dx)+U1*(xf0+dx)-Pzx*x0;
          -U0*(xf0-dx)-diag(U1)*(Pfpx*x0-dx)+Pzx*x0];        
end
% if (diry>0)
% %     U0*(yf0 - a) + U1*(yf0 - a) < Pzx*y0 + Pzu*uy < U0*(yf0 + b) + U1*(Pfpx*y0 + Pfpu*uy + b);
%     Ay = [ Pzu-diag(U1)*Pfpu;
%           -Pzu];
%     by = [ U0*(yf0+dy)+diag(U1)*(Pfpx*y0+dy)-Pzx*y0;
%           -U0*(yf0-dy)-U1*(yf0-dy)+Pzx*y0 ];  
% else    
% %     U0*(yf0 - a) + U1*(Pfpx*y0 + Pfpu*uy - a) < Pzx*y0 + Pzu*uy < U0*(yf0 + b) + U1*(yf0 + b);
%     Ay = [ Pzu;
%           -Pzu+diag(U1)*Pfpu];
%     by = [ U0*(yf0+dy)+U1*(yf0+dy)-Pzx*y0;
%           -U0*(yf0-dy)-diag(U1)*(Pfpx*y0-dy)+Pzx*y0];   
% end

% % foot velocity constraints
% Ax = [Ax;
%       Us*Pfpu;
%      -Us*Pfpu];
% bx = [bx;
%      -Us*Pfpx*x0+xs0+Tstep*vxmax;
%       Us*Pfpx*x0-xs0+Tstep*vxmax];

% Constraints on maximum velocity of foot
Ax = [Ax;
      Pfvu;
     -Pfvu];
bx = [bx;
     -Pfvx*x0+vxmax;
      Pfvx*x0+vxmax];

% % Constraints on maximum acceleration of foot wrt body
% Ax = [Ax;
%       Pfau;
%      -Pfau];
% bx = [bx;
%      -Pfax*x0+afmax;
%       Pfax*x0+afmax];

% Constraints on maximum force on foot
Ax = [Ax;
      Uf(1:Nstep+1,:);
     -Uf(1:Nstep+1,:)];
bx = [bx;
      ones(Nstep+1,1)*ufmax;
      ones(Nstep+1,1)*ufmax];  
  

% Ay = [Ay;Us*Pfpu;-Us*Pfpu];
% by = [by;-Us*Pfpx*y0+ys0+Tstep*vymax;Us*Pfpx*y0-ys0+Tstep*vymax];
%    
% A = blkdiag(Ax,Ay);
A = Ax;
% b = [bx;by];
b = bx;
% 
% % ZMP double support approximate constraints
% uxy0 = reshape(u0,N+1,2);
% xfu = uxy0(N+1:end,1);
% yfu = uxy0(N+1:end,2);
% 
% dp0 = [xfu-xf0;yfu-yf0;0];
% % constraints based on unconstrained 
% z10 = [xf0+dx;yf0-sign(dp0(1)*dp0(2))*dy;0];
% z20 = [xf0-dx;yf0+sign(dp0(1)*dp0(2))*dy;0];
% % constraints based on unconstrained 
% z1 = [xf0+dx;yf0-sign(dp0(1)*dp0(2))*dy;0];
% z2 = [xf0-dx;yf0+sign(dp0(1)*dp0(2))*dy;0];
% A = [A; dp0(2)*Pzu,zeros(N,1),-dp0(1)*Pzu,zeros(N,1);
%        -dp0(2)*Pzu,zeros(N,1), dp0(1)*Pzu,zeros(N,1)];
% b = [b;-dp0(1)*(z10(2)-Pzx*y0)+dp0(2)*(z10(1)-Pzx*x0);
%         dp0(1)*(z20(2)-Pzx*y0)-dp0(2)*(z20(1)-Pzx*x0)];

% end with zero velocity and acceleration
% Aeq = blkdiag([Us*Pfvu;Us*Pfau],[Us*Pfvu;Us*Pfau]);
% beq = [-Us*Pfvx*x0;
%        -Us*Pfax*x0;
%        -Us*Pfvx*y0
%        -Us*Pfax*y0;];

% end with zero velocity
% Aeq = blkdiag(Us*Pfvu,Us*Pfvu);
% beq = [-Us*Pfvx*x0;
%        -Us*Pfvx*y0];   
Aeq = Us*Pfvu;
beq = -Us*Pfvx*x0;

   
% end with a specific foot location
% Aeq = [Aeq;blkdiag(Us*Pfpu,Us*Pfpu)];
% beq = [beq;0.1-Us*Pfpx*x0;
%            0.1-Us*Pfpx*y0];   

%%

[ux,fval] = quadprog(Q,f,A,b,Aeq,beq,[],[],u0);    
fval

Xf = Pfpx*x0 + Pfpu*ux;

xf = Us*Xf;